/* dstralg.h - This is the API for libdstralg. */
/* Copyright (C) 2013 Robin Miyagi
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef ____DSTRALG_H____
#define ____DSTRALG_H____

/* Data types used by the dstralg library.......................... */

/* The data type used to represent a heap. */
struct dstralg_heap_struct;
typedef struct dstralg_heap_struct *dstralg_heap_T;

/* The data type used to represent a binary tree. */
struct dstralg_tree_struct;
typedef struct dstralg_tree_struct *dstralg_tree_T;

/* Type for a pointer to a comparison function. */
typedef int (*dstralg_compare_T) (const void *, const void *);

/* Type for a pointer to a destructor function. */
typedef void (*dstralg_destruct_T) (void *);

/* Interface definitions for the heap.............................. */

/* Create a new heap object, with comparison function COMPARE and
   destructor function DESTRUCT. */
extern dstralg_heap_T dstralg_heap_new (dstralg_compare_T compare,
					dstralg_destruct_T destruct);

/* Add DATA to HEAP. Returns -1 on error (ENOMEM) and 0 on success. */
extern int dstralg_heap_put (dstralg_heap_T heap, const void *data);

/* Get the smallest value from HEAP.  Returns 0 if HEAP is empty, -1
   on error (ENOMEM), and 1 on success. If there is data available on
   the heap, a pointer to the data will be returned in DATA. */
extern int dstralg_heap_get (dstralg_heap_T heap, const void **data);

/* This is the destructor function for the entire heap.  It frees all
   memory associated with the heap, including all data and nodes. */
extern void dstralg_heap_delete (dstralg_heap_T heap);

/* Interface definitions for the tree.............................. */

/* Return a newly allocated tree with a pointer to a comparison
   function passed in COMPARE, a pointer to a destructor function
   passed in DESTRUCT. */
extern dstralg_tree_T dstralg_tree_new (dstralg_compare_T compare,
					dstralg_destruct_T destruct);

/* Delete the tree specified by TREE, including all data and
   nodes. */
extern void dstralg_tree_delete (dstralg_tree_T tree);

/* Create a node containing DATA, and add to TREE.  This function
   returns 1 if TREE already contains DATA, -1 if there was an error
   (errno set to ENOMEM), and 0 if DATA was successfully added to
   TREE. */
extern int dstralg_tree_insert (dstralg_tree_T tree, const void *data);

/* Remove the data matching QUERY from TREE. Returns 0 if the tree is
   empty, -1 if the node was not found even though the tree is not
   empty, and 1 if the node was successfully removed from the tree. */
extern int dstralg_tree_remove (dstralg_tree_T tree, const void *query);

/* Searches in tree TREE for data matching QUERY and returns a pointer
   to the matching data if such data is found, or NULL if such data
   could not be found. */
extern void *dstralg_tree_query (dstralg_tree_T tree, const void *query);

/* Return a pointer to data that has the smallest value in TREE, or
   NULL if TREE is empty. */
extern void *dstralg_tree_min (dstralg_tree_T tree);

/* Return a pointer to data that has the largest value in TREE, or
   NULL if TREE is empty. */
extern void *dstralg_tree_max (dstralg_tree_T tree);

/* Return a pointer to data in TREE that has the smallest value
   greater than QUERY, or NULL if either there is no data greater than
   QUERY, there is no data matching QUERY, or TREE is empty. */
extern void *dstralg_tree_next (dstralg_tree_T tree, const void *query);

/* Return a pointer to data in TREE that has the largest value
   less than QUERY, or NULL if either there is no data less than
   QUERY, there is no data matching QUERY, or TREE is empty. */
extern void *dstralg_tree_previous (dstralg_tree_T tree, const void *query);

/* Interface definitions for the heap. */

struct dstralg_stack_struct;
typedef struct dstralg_stack_struct *dstralg_stack_T;

extern dstralg_stack_T dstralg_stack_new (dstralg_destruct_T destruct);
extern void dstralg_stack_delete (dstralg_stack_T stack);
extern int dstralg_stack_push (dstralg_stack_T stack, const void *data);
extern void *dstralg_stack_pop (dstralg_stack_T stack);

#endif	/* ! ____DSTRALG_H____ */
