/* heap.c - implementation of the heap data structure and
   corresponding algorithms. */
/* Copyright (C) 2013 Robin Miyagi
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Include the API header file(s) for debugging and for use from
   within this, the file including the API header file(s). */
#include <dstralg.h>

/* Package configuration header. */
#include <config.h>

/* Project includes. */
#include <include/project.h>

/* System includes. */
#include <stddef.h>
#include <stdlib.h>

/* Macro definitions for calculating the left/right element of
   cursor. */
#define get_left_cursor(cursor) (2 * (cursor) + 1)
#define get_right_cursor(cursor) (2 * (cursor) + 2)

struct dstralg_heap_struct
{
  size_t last;
  const void **vector;
  dstralg_compare_T compare;
  dstralg_destruct_T destruct;
};
typedef struct dstralg_heap_struct *dstralg_heap_T;

/* FORWARD DECLARATIONS ........................................... */
dstralg_heap_T dstralg_heap_new (dstralg_compare_T compare, dstralg_destruct_T destruct);
int dstralg_heap_put (dstralg_heap_T heap, const void *data);
int dstralg_heap_get (dstralg_heap_T heap, const void **data);
void dstralg_heap_delete (dstralg_heap_T heap);

static int dstralg_heap_increase (dstralg_heap_T heap);
static void normalize_after_put (dstralg_heap_T heap, size_t cursor);
static size_t get_parent_cursor (size_t cursor);
static int dstralg_heap_decrease (dstralg_heap_T heap);
static void normalize_after_get (dstralg_heap_T heap, size_t cursor);

/* IMPLEMENTATIONS ................................................ */
/* Create a new heap object, with comparison function COMPARE and
   destructor function DESTRUCT. */
dstralg_heap_T
dstralg_heap_new (dstralg_compare_T compare, dstralg_destruct_T destruct)
{
  dstralg_heap_T heap;

  heap = calloc (1, sizeof (struct dstralg_heap_struct));
  if (heap == NULL)
    return NULL;
  heap->compare = compare;
  heap->destruct = destruct;
  return heap;
}
/* Add DATA to HEAP. Returns -1 on error (ENOMEM) and 0 on success. */
int
dstralg_heap_put (dstralg_heap_T heap, const void *data)
{
  if (dstralg_heap_increase (heap) == -1)
    return -1;
  heap->vector[heap->last] = data;
  normalize_after_put (heap, heap->last);
  ++(heap->last);
  return 0;
}
/* Get the smallest value from HEAP.  Returns 0 if HEAP is empty, -1
   on error (ENOMEM), and 1 on success. If there is data available on
   the heap, a pointer to the data will be returned in DATA. */
int
dstralg_heap_get (dstralg_heap_T heap, const void **data)
{
  if (heap->vector == NULL)
    return 0;
  *data = heap->vector[0];
  --(heap->last);
  heap->vector[0] = heap->vector[heap->last];
  if (dstralg_heap_decrease (heap) == -1)
    return -1;
  normalize_after_get (heap, 0);
  return 1;
}
/* This is the destructor function for the entire heap.  It frees all
   memory associated with the heap, including all data and nodes. */
void
dstralg_heap_delete (dstralg_heap_T heap)
{
  size_t i;

  for (i = 0; i < heap->last; ++i)
    heap->destruct ((void *) heap->vector[i]);
  free (heap->vector);
  free (heap);
}
static int
dstralg_heap_increase (dstralg_heap_T heap)
{
  heap->vector = realloc (heap->vector, (heap->last + 1) * sizeof (void *));
  if (heap->vector == NULL)
    return -1;
  else
    return 0;
}
static void
normalize_after_put (dstralg_heap_T heap, size_t cursor)
{
  size_t parent;
  const void **cursor_ptr;
  const void **parent_ptr;

  if (cursor == 0)
    return;
  parent = get_parent_cursor (cursor);
  cursor_ptr = &(heap->vector[cursor]);
  parent_ptr = &(heap->vector[parent]);
  if (heap->compare (*cursor_ptr, *parent_ptr) < 0)
    swap (cursor_ptr, parent_ptr);
  else
    return;
  normalize_after_put (heap, parent);
}
static size_t
get_parent_cursor (size_t cursor)
{
  size_t parent;
  size_t cursor_level;
  size_t parent_level;
  size_t cursor_offset;
  size_t parent_offset;
  size_t level;

  for (level = 0, cursor_level = 0; cursor_level <= cursor; ++level)
    cursor_level += (1 << level);
  --level;
  cursor_level -= (1 << level);
  parent_level = cursor_level - (1 << (level - 1));
  cursor_offset = cursor - cursor_level;
  parent_offset = cursor_offset / 2;
  parent = parent_level + parent_offset;
  return parent;
}

static int
dstralg_heap_decrease (dstralg_heap_T heap)
{
  heap->vector = realloc (heap->vector, (heap->last) * sizeof (void *));
  if ((heap->vector == NULL) && (heap->last != 0))
    return -1;
  return 0;
}
static void
normalize_after_get (dstralg_heap_T heap, size_t cursor)
{
  size_t left;
  size_t right;
  size_t smaller;
  const void **cursor_ptr;
  const void **left_ptr;
  const void **right_ptr;
  const void **smaller_ptr;

  left = get_left_cursor (cursor);
  right = get_right_cursor (cursor);
  cursor_ptr = &(heap->vector[cursor]);
  left_ptr = &(heap->vector[left]);
  right_ptr = &(heap->vector[right]);
  if ((left >= heap->last) && (right >= heap->last))
    return;
  else if (left >= heap->last)
    smaller = right;
  else if (right >= heap->last)
    smaller = left;
  else
    {
      if (heap->compare (*right_ptr, *left_ptr) < 0)
	smaller = right;
      else
	smaller = left;
    }
  smaller_ptr = &(heap->vector[smaller]);
  if (heap->compare (*cursor_ptr, *smaller_ptr) > 0)
    swap (cursor_ptr, smaller_ptr);
  else
    return;
  normalize_after_get (heap, smaller);
}
