/* tree.c - implementation of the AVL tree data structure and
   associated functions. */
/* Copyright (C) 2013 Robin Miyagi
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Package configuration header. */
#include <config.h>

/* API header file(s). */
#include <dstralg.h>

/* Project headers. */
#include <include/project.h>

/* System headers. */
#include <stdlib.h>
#include <stddef.h>

/* Balance factors. */
#define LEFT_UNBALANCED (-2)
#define LEFT_HEAVY (-1)
#define BALANCED 0
#define RIGHT_HEAVY 1
#define RIGHT_UNBALANCED 2

struct node_struct
{
  int balance;
  struct node_struct *parent;
  struct node_struct *left;
  struct node_struct *right;
  const void *data;
};
typedef struct node_struct node_T;

struct dstralg_tree_struct
{
  node_T *root;
  dstralg_compare_T compare;
  dstralg_destruct_T destruct;
};
typedef struct dstralg_tree_struct *dstralg_tree_T;

/* FORWARD DECLARATIONS ........................................... */
dstralg_tree_T dstralg_tree_new (dstralg_compare_T compare, dstralg_destruct_T destruct);
void dstralg_tree_delete (dstralg_tree_T tree);
int dstralg_tree_insert (dstralg_tree_T tree, const void *data);
int dstralg_tree_remove (dstralg_tree_T tree, const void *query);
void *dstralg_tree_query (dstralg_tree_T tree, const void *query);
void *dstralg_tree_min (dstralg_tree_T tree);
void *dstralg_tree_max (dstralg_tree_T tree);
void *dstralg_tree_next (dstralg_tree_T tree, const void *query);
void *dstralg_tree_previous (dstralg_tree_T tree, const void *query);
static void normalize_after_insert (dstralg_tree_T tree, node_T *node);
static void normalize_after_remove (dstralg_tree_T tree, node_T *node);
static int rebalance_tree (dstralg_tree_T tree, node_T *cursor, node_T *parent,
			   node_T *grandparent);

/* generic binary tree functions................................... */
static int bintree_search (dstralg_tree_T tree, const void *query,
			   node_T **result);
static void rotate_left (dstralg_tree_T tree, node_T *cursor);
static void rotate_right (dstralg_tree_T tree, node_T *cursor);
static node_T *get_max_node (node_T *cursor);
static node_T *get_min_node (node_T *cursor);
static void delete_tree (dstralg_tree_T tree, node_T *node);

/* IMPLEMENTATIONS ................................................ */
/* Return a newly allocated tree with a pointer to a comparison
   function passed in COMPARE, a pointer to a destructor function
   passed in DESTRUCT. */
dstralg_tree_T
dstralg_tree_new (dstralg_compare_T compare, dstralg_destruct_T destruct)
{
  dstralg_tree_T tree;

  tree = calloc (1, sizeof (struct dstralg_tree_struct));
  if (tree == NULL)
    return NULL;
  tree->compare = compare;
  tree->destruct = destruct;
  return tree;
}
/* Delete the tree specified by TREE, including all data and
   nodes. */
void
dstralg_tree_delete (dstralg_tree_T tree)
{
  delete_tree (tree, tree->root);
  free (tree);
}
int
dstralg_tree_insert (dstralg_tree_T tree, const void *data)
{
  node_T *node;
  node_T *parent;
  
  if (bintree_search (tree, data, &parent) == 1)
    return 1;
  node = malloc (sizeof (node_T));
  if (node == NULL)
    return -1;
  node->balance = BALANCED;
  node->left = NULL;
  node->right = NULL;
  node->parent = parent;
  node->data = data;
  if (parent == NULL)
    tree->root = node;
  else if (tree->compare (node->data, parent->data) < 0)
    parent->left = node;
  else
    parent->right = node;
  normalize_after_insert (tree, node);
  return 0;
}
int
dstralg_tree_remove (dstralg_tree_T tree, const void *query)
{
  node_T *parent;	/* parent of delete_me */
  node_T *delete_me;
  node_T *cursor;

  if (tree->root == NULL)
    return 0;
  if (bintree_search (tree, query, &cursor) == 0)
    return -1;
  delete_me = get_max_node (cursor->left);
  if (delete_me == NULL)
    delete_me = get_min_node (cursor->right);
  if (delete_me == NULL)
    delete_me = cursor;
  tree->destruct ((void *) cursor->data);
  if (delete_me != cursor)
    cursor->data = delete_me->data;
  parent = delete_me->parent;
  if (parent == NULL)
    {
      if (delete_me->left != NULL)
	tree->root = delete_me->left;
      else if (delete_me->right != NULL)
	tree->root = delete_me->right;
      else
        tree->root = NULL;
    }
  else if (delete_me == parent->left)
    {
      if (delete_me->left != NULL)
	parent->left = delete_me->left;
      else if (delete_me->right != NULL)
	parent->left = delete_me->right;
      else
	parent->left = NULL;
      ++(parent->balance);
    }
  else
    {
      if (delete_me->left != NULL)
	parent->right = delete_me->left;
      else if (delete_me->right != NULL)
	parent->right = delete_me->right;
      else
	parent->right = NULL;
      --(parent->balance);
    }
  free (delete_me);
  normalize_after_remove (tree, parent);
  return 1;
}
/* Searches in tree TREE for data matching QUERY and returns a pointer
   to the matching data if such data is found, or NULL if such data
   could not be found. */
void
*dstralg_tree_query (dstralg_tree_T tree, const void *query)
{
  node_T *node;

  if (bintree_search (tree, query, &node) == 0)
    return NULL;
  return (void *) node->data;
}
/* Return a pointer to data that has the smallest value in TREE, or
   NULL if TREE is empty. */
void
*dstralg_tree_min (dstralg_tree_T tree)
{
  node_T *node;

  node = get_min_node (tree->root);
  return (void *) node->data;
}
/* Return a pointer to data that has the largest value in TREE, or
   NULL if TREE is empty. */
void
*dstralg_tree_max (dstralg_tree_T tree)
{
  node_T *node;

  node = get_max_node (tree->root);
  return (void *) node->data;
}
/* Return a pointer to data in TREE that has the smallest value
   greater than QUERY, or NULL if either there is no data greater than
   QUERY, there is no data matching QUERY, or TREE is empty. */
void
*dstralg_tree_next (dstralg_tree_T tree, const void *query)
{
  node_T *match;
  node_T *ptr;
  node_T *next;
  
  if (bintree_search (tree, query, &match) != 1)
    return NULL;
  next = get_min_node (match->right);
  if (next == NULL)
    {
      for (ptr = match->parent; ptr != NULL; ptr = ptr->parent)
	{
	  if (ptr->left == match)
	    {
	      next = ptr;
	      break;
	    }
	  match = ptr;
	}
      if (ptr == NULL)
	return NULL;
    }
  return (void *) next->data;
}
/* Return a pointer to data in TREE that has the largest value
   less than QUERY, or NULL if either there is no data less than
   QUERY, there is no data matching QUERY, or TREE is empty. */
void
*dstralg_tree_previous (dstralg_tree_T tree, const void *query)
{
  node_T *match;
  node_T *ptr;
  node_T *previous;

  if (bintree_search (tree, query, &match) != 1)
    return NULL;
  previous = get_max_node (match->left);
  if (previous == NULL)
    {
      for (ptr = match->parent; ptr != NULL; ptr = ptr->parent)
	{
	  if (ptr->right == match)
	    {
	      previous = ptr;
	      break;
	    }
	  match = ptr;
	}
      if (ptr == NULL)
	return NULL;
    }
  return (void *) previous->data;
}
static void
normalize_after_insert (dstralg_tree_T tree, node_T *node)
{
  node_T *parent = NULL;
  node_T *grandparent = NULL;
  node_T *cursor;
  node_T *ptr;
  
  for (ptr = node; ptr != NULL; ptr = ptr->parent)
    {
      cursor = parent;
      parent = grandparent;
      grandparent = ptr;
      if (parent != NULL)
	{
	  if (grandparent->left == parent)
	    --(grandparent->balance);
	  else
	    ++(grandparent->balance);
	  if (grandparent->balance == BALANCED)
	    return;
	}
      if ((grandparent->balance == LEFT_UNBALANCED)
	  || (grandparent->balance == RIGHT_UNBALANCED))
	break;
    }
  if (ptr == NULL)
    return;
  rebalance_tree (tree, cursor, parent, grandparent);
}
static void
normalize_after_remove (dstralg_tree_T tree, node_T *node)
{
  node_T *previous;
  node_T *ptr;
  node_T *grandparent = NULL;
  node_T *parent;
  node_T *cursor;
  int balance_result;
  
  for (ptr = node; ptr != NULL; ptr = ptr->parent)
    {
      previous = grandparent;
      grandparent = ptr;
      if (previous != NULL)
	{
	  if (grandparent->left == previous)
	    ++(grandparent->balance);
	  else
	    --(grandparent->balance);
	}
      if ((grandparent->balance == LEFT_HEAVY)
	  || (grandparent->balance == RIGHT_HEAVY))
	return;
      if (grandparent->balance == LEFT_UNBALANCED)
	{
	  parent = grandparent->left;
	  break;
	}
      if (grandparent->balance == RIGHT_UNBALANCED)
	{
	  parent = grandparent->right;
	  break;
	}
    }
  if (ptr == NULL)
    return;
  if (parent->balance == LEFT_HEAVY)
    cursor = parent->left;
  else
    cursor = parent->right;
  balance_result = rebalance_tree (tree, cursor, parent, grandparent);
  if (balance_result == 0)
    normalize_after_remove (tree, parent);
  else if (balance_result == 1)
    normalize_after_remove (tree, cursor);
}
static int
rebalance_tree (dstralg_tree_T tree, node_T *cursor, node_T *parent,
		node_T *grandparent)
{
  int retval;
  
  if (grandparent->balance == LEFT_UNBALANCED)
    {
      if (parent->balance == BALANCED)
	{
	  rotate_right (tree, grandparent);
	  parent->balance = RIGHT_HEAVY;
	  grandparent->balance = LEFT_HEAVY;
	  retval = -1;
	}
      else if (parent->balance == LEFT_HEAVY)
	{
	  rotate_right (tree, grandparent);
	  parent->balance = BALANCED;
	  grandparent->balance = BALANCED;
	  retval = 0;
	}
      else
	{
	  rotate_left (tree, parent);
	  rotate_right (tree, grandparent);
	  if (cursor->balance == BALANCED)
	    {
	      parent->balance = BALANCED;
	      grandparent->balance = BALANCED;
	    }
	  else if (cursor->balance == LEFT_HEAVY)
	    {
	      parent->balance = BALANCED;
	      grandparent->balance = RIGHT_HEAVY;
	    }
	  else
	    {
	      parent->balance = LEFT_HEAVY;
	      grandparent->balance = BALANCED;
	    }
	  cursor->balance = BALANCED;
	  retval = 1;
	}
    }
  else
    {
      /* Else the grandparent is right-unbalanced. */
      if (parent->balance == BALANCED)
	{
	  rotate_left (tree, grandparent);
	  parent->balance = LEFT_HEAVY;
	  grandparent->balance = RIGHT_HEAVY;
	  retval = -1;
	}
      else if (parent->balance == RIGHT_HEAVY)
	{
	  rotate_left (tree, grandparent);
	  parent->balance = BALANCED;
	  grandparent->balance = BALANCED;
	  retval = 0;
	}
      else
	{
	  rotate_right (tree, parent);
	  rotate_left (tree, grandparent);
	  if (cursor->balance == BALANCED)
	    {
	      parent->balance = BALANCED;
	      grandparent->balance = BALANCED;
	    }
	  else if (cursor->balance == LEFT_HEAVY)
	    {
	      parent->balance = RIGHT_HEAVY;
	      grandparent->balance = BALANCED;
	    }
	  else
	    {
	      parent->balance = BALANCED;
	      grandparent->balance = LEFT_HEAVY;
	    }
	  cursor->balance = BALANCED;
	  retval = 1;
	}
    }
  return retval;
}
/* generic binary tree functions................................... */

/* Do a binary tree search in TREE for a node whose data matches
   QUERY. If a matching result was found this function returns 1, and
   RESULT contains a pointer to the matching node.  If a node was not
   found, this function returns 0 and RESULT contains either NULL
   (empty tree), or the node that would have been the parent node in
   matching case. */
static int
bintree_search (dstralg_tree_T tree, const void *query, node_T **result)
{
  node_T *ptr;
  int compare;
  *result = NULL;
  for (ptr = tree->root; ptr != NULL;)
    {
      *result = ptr;
      compare = tree->compare (query, ptr->data);
      if (compare == 0)
	return 1;

      /* Loop update operation. */
      if (compare < 0)
	ptr = ptr->left;
      else
	ptr = ptr->right;
    }
  return 0;
}
/* Rotate the sub-tree with root CURSOR counter-clockwise (left) about
   CURSOR. */
static void
rotate_left (dstralg_tree_T tree, node_T *cursor)
{
  node_T *right;
  node_T *parent;
  node_T *sub_left;

  right = cursor->right;
  parent = cursor->parent;
  sub_left = right->left;
  if (parent == NULL)
    tree->root = right;
  else
    {
      if (parent->left == cursor)
	parent->left = right;
      else
	parent->right = right;
    }
  right->parent = parent;
  right->left = cursor;
  cursor->parent = right;
  cursor->right = sub_left;
  if (sub_left != NULL)
    sub_left->parent = cursor;
}
/* Rotate the sub-tree with root CURSOR clockwise (right) about
   CURSOR. */
static void
rotate_right (dstralg_tree_T tree, node_T *cursor)
{
  node_T *left;
  node_T *parent;
  node_T *sub_right;

  left = cursor->left;
  parent = cursor->parent;
  sub_right = left->right;
  if (parent == NULL)
    tree->root = left;
  else
    {
      if (parent->left == cursor)
	parent->left = left;
      else
	parent->right = left;
    }
  left->parent = parent;
  left->right = cursor;
  cursor->parent = left;
  cursor->left = sub_right;
  if (sub_right != NULL)
    sub_right->parent = cursor;
}
/* Return the node that has data greater than the data of CURSOR,
   CURSOR if there is no node with data greater than that of CURSOR,
   or NULL if cursor is NULL. */
static node_T
*get_max_node (node_T *cursor)
{
  node_T *max_node = NULL;
  node_T *ptr;

  for (ptr = cursor; ptr != NULL; ptr = ptr->right)
    max_node = ptr;
  return max_node;
}
/* Return the node with data less than the data of CURSOR, CURSOR if
   there is no node with data less than that of CURSOR, or NULL if
   CURSOR was NULL. */
static node_T
*get_min_node (node_T *cursor)
{
  node_T *min_node = NULL;
  node_T *ptr;

  for (ptr = cursor; ptr != NULL; ptr = ptr->left)
    min_node = ptr;
  return min_node;
}
/* Delete the entire sub-tree rooted at NODE.  If NODE is the root of
   the entire tree, then the entire tree will be deleted. */
static void
delete_tree (dstralg_tree_T tree, node_T *node)
{
  if (node == NULL)
    return;
  delete_tree (tree, node->left);
  delete_tree (tree, node->right);
  tree->destruct ((void *) node->data);
  free (node);
}
